#include "flexibleMRAITimer.h"
#include "../rtProtoBGP.h"


  // ----- constructor flexibleMRAITimer ------------------------------------ //
  /**

   *
   * @param b    The rtProtoBGP with which this timer is associated.

   */
FlexibleMRAITimer::FlexibleMRAITimer ( rtProtoBGP * b ) 
	:bgp( b ), offset( 0 )
{
}

  // ----- flexibleMRAITimer.expire --------------------------------------- //
  /**

   */
void
FlexibleMRAITimer::expire ( Event * e ) {
  int prefix = 0;
  set < unsigned int > current_prefixes ;// prefixes for which this timer is used 
  //when flexibleMRAITimer is used then value for each prefix depends upon timer_type
  //  timer_type is of two types: fast and slow
  // Events Tlong and Tdown cames under slow category
  // Events Tshort and Tup cames under fast category
 // If timer_type is slow then timer value for respective prefix is 30 sec
  // and if timer_type is fast then value for respective prefix is 15 sec.
  float next_timer ;

  bgp -> expired_last = offset ;
 //expired_last contains offset, i.e. the number of the timer that expired the last
 
 set < unsigned int > :: iterator prefix_iter ;
  int new_scheduled_time = FAST;
  for ( prefix_iter =  prefixes . begin ( ) ; prefix_iter != prefixes . end ( ) ; prefix_iter ++ )
  {
    prefix = *prefix_iter ;

    if ( bgp -> now ( ) - bgp -> lastUpdate [ prefix ] > bgp -> idle_time [ prefix ] )
    {
      bgp -> idle_time [ prefix ] = bgp -> now ( ) - bgp -> lastUpdate [ prefix ];
    }
    //I had errors calculating idle_time. What if it the last message received in the prevous round?
    //An idle_time should be calculated from the begining of the new round?
    //What if the offset timer didn't expire, than idle_time cannot be longer than flexibleMRAI, so
    //expiry of a timer is treated as receiving a msg
    bgp -> lastUpdate [ prefix ] = bgp -> now ( );

	if ( ! bgp -> flexible_MRAI ) {
		next_timer =  offset ;
	} else {
		//RG changes to reusable MRAI timers:
		  if ( bgp -> idle_time [ prefix ] < 1 ) {
//			  bgp -> reusable_timer[ bgp -> prefix_timer[ prefix ] ] -> event_type = rtProtoBGP::TLONG;
			//If the idle time is less than 1 sec then timer is of 30 sec
			//which means it comes under category of "slow" in timer_type event  
			
			bgp -> flexibleMRAI [ prefix ] = SLOW;
//              next_timer = offset + (int)SLOW;
			 
			 bgp -> granularity = 2;
			//since prefix_time = number of timer*granularity; number of timers are fixed ie 15;
			// For fast, I want to assign 30 sec MRAI round. So, granularity is 2 sec for fast category.

		  } else {
//			  bgp -> reusable_timer[ bgp -> prefix_timer[ prefix ] ] -> event_type = rtProtoBGP::TSHORT;
			  bgp -> flexibleMRAI [ prefix ] = FAST;
			//If the idle time is more than 1 sec then timer is of 15 sec
			//which means it comes under category of "fast" in timer_type event  
//			  next_timer = offset + (int) FAST;
			  bgp -> granularity = 1;
			//since prefix_time = number of timer*granularity; number of timers are fixed ie 15;
			// For slow, I want to assign 15 sec MRAI round. So, granularity is 1 sec for slow category.

		  }
		  next_timer = offset + (int) ( bgp -> flexibleMRAI[ prefix ] );
      if ( next_timer >= 30 )
      {
        next_timer -= 30 ;
      }
	}
	//the next_timer is not determined (there is no timer, i.e.MRAI round will be 30 sec), but if there are msgs to be sent,
    //the next_timer will be recalculated. If no msg is sent than next_timer will be calculated when a msg is sent from new_try_update_sent
    bgp -> prefix_timer [ prefix ] = -1; //not yet associated with any message/peer

	  bgp -> round[ prefix ] ++;
	  bgp -> idle_time[ prefix ] = 0;
  }

  //this round is over, so BGP has to erase all prefixes
  //Of course, if there are msgs to be sent, prefixes will be added 
  prefixes . clear ( ) ;
  //for each peer I first check which prefixes use this timer. These prefixes are put in the current_prefixes, and also deleted
  //from peer_timer.
  //Secondly, I check are there any waiting_adv for the prefixes in the current_prefixes. The waiting_adv are sent, and their prefixes
  //are placed in peer_timer (for these prefixes timer is used again)
  for (int i = 0 ; i < bgp -> nbs_ . size ( ) ; i++ )
  {
    PeerEntry * peer = bgp -> nbs_[i];

    //Maybe I don't need peer_prefix at all. Without them I will have to go through all elements of the waiting_adv
    //Who knows, maybe that is better solution, especial for small number of prefixes, for highly connected networks
    //For now I will leave the solution with peer_prefix because it seems to be more general

    //copy all prefixes for this peer in the current_prefixes
    current_prefixes = peer_prefix [ peer ] ;
    //we use current prefixes for sending msg from waiting_adv, and we will add newely send prefixes in peer_prefix
    peer_prefix [ peer ] . clear ( ) ;
                                                                                                                                                                   
    map < unsigned int, AdvPair > :: iterator i ;
    set < unsigned int > :: iterator curr_prefix ;
    for ( curr_prefix = current_prefixes . begin ( ) ; curr_prefix != current_prefixes . end ( ) ; curr_prefix ++ )
    {
    
      i = peer -> waiting_adv . find ( *curr_prefix ) ;

      prefix = *curr_prefix ;

      if ( i != peer -> waiting_adv . end ( ) )
      {
          AdvPair pair =  i -> second;
          Route * waitingrte = pair.rte;
          if ( waitingrte == NULL)
          {
            peer->waiting_adv.erase( i );
            continue;
          }
        
          // advertise the waiting route
          UpdateMessage * upmsg = new UpdateMessage(bgp->nh, waitingrte);

          bgp->send(upmsg, peer);
          bgp->reset_timer(peer, BGP_Timer::KEEPALIVE);
          peer->waiting_adv.erase( i );
		  
		  bgp -> reusable_timer [ next_timer ] -> peer_prefix [ peer ] . insert ( prefix ) ;
          bgp -> reusable_timer [ next_timer ] -> prefixes . insert ( prefix ) ;
 	//insert current message prefix to reusable_timer 
	// make association between reusable timer and message/peer
          //next_timer that is going to be used for this prefix
          bgp -> prefix_timer [ prefix ] = next_timer ;

      }
    }
	//end of processing for this peer, we need to erase all prefixes from current
    current_prefixes . clear ( ) ;
  } //end of peers
  resched ( bgp -> number_of_timers * bgp -> granularity ); //new_scheduled_time
//reschedule reusable timer =  number_of_timers * granularity
}
