#ifndef FLEXIBLE_MRAI_TIMER
#define FLEXIBLE_MRAI_TIMER

#include "timer-handler.h"
#include "set" 
#include "iostream" 
#include "../route.h"
#include "../global.h"
#include "../Comm/updatemessage.h"
#include "../peer-entry.h" 

                                  
/**
 * Author: Rajvir Gill <rajvirg@sfu.ca>
 */
class rtProtoBGP;
class PeerEntry;

//RG
class FlexibleMRAITimer : public TimerHandler {
	
	public:
		enum TIMER_TYPE {FAST = 15, SLOW = 30, NONE = -1};
		//fast includes Tshort and Tup events
		//slow includes Tlong and Tdown events

		FlexibleMRAITimer(rtProtoBGP * b);
		~FlexibleMRAITimer(){}
		
		//offset is the serial number of the timer
	  	//offset of timer describes the postion of this timer among all other offset timers
	  	//e.g. offset=0 is the first timer, offset=1 is the timer that starts 1 sec after the first one
		// timer_type describes the type ie. whether it is fast or slow
		// timer is scheduled according to timer_type 
		// when timer expires, it is rescheduled back 
		void reschedule(TIMER_TYPE type = FAST) { resched( (float) type ); scheduled_time = (TIMER_TYPE) offset; }
		void sched( float offset ) { TimerHandler::sched( offset ); scheduled_time = (TIMER_TYPE) offset; }

		void expire(Event *e);
		void set_offset ( float input_offset ) { offset = input_offset; }

	public:
		 /** The rtProtoBGP with which this flexible algorithm is associated. */
	  rtProtoBGP *bgp;

	  //a map keyed by peers, where the values are vecotors of prefixes
	  //for the each peer (key) the map contains vector (array) of prefix associated withe them
	  //for example, when a timer expires, it will send updates about to peer 1 containing following 3 prefixes, (prefix 1, prefix 2, prefix 3)
	  map < PeerEntry*, set < unsigned int > > peer_prefix ;

	  //this set contains all prefexes that use this timer
	  set < unsigned int > prefixes ;

	public:

	  //offset is the serial number of the timer
	  //offset of timer describes the postion of this timer among all other offset timers
	  //e.g. offset=0 is the first timer, offset=1 is the timer that starts 1 sec after the first one
	  float offset;

	  //RG
	  //This describes event type: whether it is Tlong/ Tshort/ Tup or Tdown
	  // Events are described in rtprotoBGP.h 
	  int event_type;

	  //RG
	  // This describes timer values for respective timer_types 
	  // i.e. whether type is fast or slow
	  // slow associates with Tlong and Tdown
	  //whereas, fast associated with Tup and Tshort
	  TIMER_TYPE scheduled_time;
};
#endif
