/*
 * Copyright (c) 2003 Communication Networks Lab, Simon Fraser University.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Authors: Steve Wei Shen <wshen@cs.sfu.ca>,  Z. Morley Mao.
 *
 */


#ifndef DAMPINFO_H
#define DAMPINFO_H

#include <string>

using namespace std;

/** Contains route flap damping information about a particular route */

class rtProtoBGP;
class RouteInfo;
class DampReuseTimer;
class IPaddress;

class DampInfo {

public:

  DampInfo(RouteInfo* ri, rtProtoBGP* b, double wp, double cp, double sl, double rl, double hl, double mst);
  void updatePenalty(bool withdraw);
  void update(bool withdraw, RouteInfo* ri);
  string toString();


  //--------------------- member data ----------------------------

 public: 

  rtProtoBGP* bgp;
  RouteInfo* routeInfo;   //anouncement routeinfo
  double penalty;
  double lastUpdatedTime;
  bool suppressed;
  int prevUpdate;     //1: announcement,  0: withdrawal
  DampReuseTimer* reuseTimer;
  int flapCount;
  int prevDop;
  int twoBits;  //00(0): undefined,  01(1): equal,  10(2): better,  11(3): worse
  string peerNh;
  IPaddress* nlri;
  double reuseLimit;
  double suppressLimit;
  static int flapNum;  //keep track of the total number of all flaps so far
  static int supNum;  //keep track of the total number of all suppresses so far

 private:

  double WDcount;   //penalty for withdrawal after announcement
  int flapWDcount;  
  double changePenalty;
  double withdrawPenalty;
  double penalty_ceiling;
  double halfLife_ok;  //half life of decay when network is up
  double halfLife_ng;  //half life of decay when network is down
  double maxSuppressTime;
};

#endif





