/*
 * Copyright (c) 2003 Communication Networks Lab, Simon Fraser University.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Authors: Steve  Wei Shen <wshen@cs.sfu.ca>,  Z. Morley Mao.
 *
 */

#include "damp_reuse_timer.h"
#include "../rtProtoBGP.h"
#include "../dampinfo.h"

/** Constructs a damping reuse timer */
DampReuseTimer::DampReuseTimer(rtProtoBGP* b, double t, DampInfo* di): TimerHandler() {
  bgp = b;
  dampInfo = di;
  supTime = t;
}

/** Called when timer expires, and updates the damped route's status */
void DampReuseTimer::expire(Event*) {
  //Run the decision process again to determine whether to use the unsuppressed route 
  //in case the suppress time is already the maximum value allowed
  if(supTime > bgp->max_suppress_time - 0.1) {
    dampInfo->penalty = 0;
    dampInfo->suppressed = false;
    dampInfo->lastUpdatedTime = bgp->now();
  } 
  else {
    dampInfo->updatePenalty(false);
    if(dampInfo->penalty <= dampInfo->reuseLimit + 0.1) {
      dampInfo->suppressed = false;
    }
    else {
      return;
    }
  }
  if(dampInfo->prevUpdate == 0) {   //still unreachable though unsuppressed
    return;
  }
  
  //Rerun the decision process, potentially announce the route to the peer
  //Run part of decision process 2 and call 3
  RouteInfo* info = dampInfo->routeInfo;
  vector<RouteInfo*> changedinfo;
  changedinfo.push_back(info);
  vector<RouteInfo*> locribchanges;
  locribchanges = bgp->decision_process_2(changedinfo, true);
  bgp->decision_process_3(locribchanges);
}

/** Resets the timer */
void DampReuseTimer::reset() {
  if (status_ == TIMER_PENDING) {
     cancel();
  }  
  status_ = TIMER_IDLE;
}

/** Shows whether the timer is already scheduled */
bool DampReuseTimer::scheduled() {
  if(status_ == TIMER_PENDING) {
    return true;
  }
  else {
    return false;
  }
}




