

#ifndef PROCESSING_DELAY_H
#define PROCESSING_DELAY_H

#include "timer-handler.h" 
#include "iostream"
#include "rng.h"
#include "../route.h"
#include "../global.h"
#include "../peer-entry.h"
#include "../Comm/updatemessage.h"

class rtProtoBGP;
 
                                                                                                   
/**
 *  CPUTimer class models CPU processing delay.
 *  From Anja's paper (PAM04) I learned that Cisco routers have timer that
 *  limits the update processing to 5 times a second (so, one round is 200ms)
 *  Therefore, in this class I simulate that processing. To introduce offeset
 *  amid routers' processing time when I start them in the rtProtoBGP.cc with
 *  random offset (between 0 and 200ms). 
 *  Furthermore, this CPUtimer alweys runs, no matter are there any update messages or not.
 *  Once started, router run deterministicly with same value of the rounds.

 * Author: Nenad Laskovic <nlaskovi@cs.sfu.ca>

 */
 
class CPUTimer : public TimerHandler  {
public:
	CPUTimer(rtProtoBGP *b);
	~CPUTimer();
  
//protected:
public:
  void expire(Event *e);
  
  /** The rtProtoBGP with which this timer is associated. */
  rtProtoBGP *bgp;

  //default value is 200ms, but in order to model higher CPU loads it could be set to higher value
  double round ;

};

#endif


