
#include "reusableMRAItimer.h"     
#include "../rtProtoBGP.h"


  // ----- constructor reusableMRAItimer ------------------------------------ //
  /**

   *
   * @param b    The rtProtoBGP with which this timer is associated.

   */
reusableMRAItimer::reusableMRAItimer(rtProtoBGP *b)  : TimerHandler()
{
  bgp = b;
  offset = 0;
}

  // ----- destructor reusableMRAItimer ------------------------------------ //
  /**

   */

reusableMRAItimer::~reusableMRAItimer() {}


  // ----- set_offset reusableMRAItimer ------------------------------------ //
  /**
   */

void reusableMRAItimer::set_offset ( float input_offset )
{
  offset = input_offset;
}

  // ----- reusableMRAItimer.expire --------------------------------------- //
  /**

   */
void reusableMRAItimer::expire(Event*)
{
  int prefix = 0;
  set < unsigned int > current_prefixes ; // prefixes for which this timer is used (different for each peer)
  //when adaptiveMRAI is used then next timer is not the timer in which we are (the one that just has expired)
  //next_timer is determined by the value of the adaptiveMRAI (Note, when adaptive is not used, that is the same
  //as that adaptive is 30 sec, so next timer is the same that just has expired:) Clear, n'est pas??
  //for example when adaptiveMRAI is 8 seconds, then after timer 10 comes timer 18 not timer 10 again as
  // in the case without adaptiveMRAI
  float next_timer ;


  bgp -> expired_last = offset ; //expired_last contains offset, i.e. the number of the timer that expired the last

  // we want to find the first time when this timer expires for current prefix mo matter for which peer, and that is IDLE time
  // expiration for the rest of peer will give the sama or fault result.
  // I'm looking for IDLE time for this prefix and all peers of one BGP speaker in current MRAI round
  //this vector contains all prefixes in one particluar peer that use this timer
  set < unsigned int > :: iterator prefix_iter ;
  for ( prefix_iter =  prefixes . begin ( ) ; prefix_iter != prefixes . end ( ) ; prefix_iter ++ )
  {
    prefix = *prefix_iter ;

    if ( bgp -> now ( ) - bgp -> lastUpdate [ prefix ] > bgp -> idle_time [ prefix ] )
    {
      bgp -> idle_time [ prefix ] = bgp -> now ( ) - bgp -> lastUpdate [ prefix ];
    }

    //I had errors calculating idle_time. What if it the last message received in the prevous round?
    //An idle_time should be calculated from the begining of the new round?
    //What if the offset timer didn't expire, than idle_time cannot be longer than adaptiveMRAI, so
    //expiry of a timer is treated as receiving a msg
    bgp -> lastUpdate [ prefix ] = bgp -> now ( );
    
    if ( ! bgp -> adaptive_MRAI )
    {
      //if adaptive algorithm is not used, than next timer is this timer itself. MRAI is 30 sec
      next_timer =  offset ;
    }
    else
    {
      float current_active, current_sigma ;
      int round = bgp -> round [ prefix ] ;
      
      //is this the first round?
      if ( round < 2 )
      {
        //the first round has just finished, adaptiveMRAI needs to be calculated
        bgp -> adaptiveMRAI [ prefix ] = bgp -> global_mrai ;
      }

      current_active = bgp -> adaptiveMRAI [ prefix ] - bgp -> idle_time [ prefix ] ;

      //if a round is skiped, than is possible that currnet_active be less than 0
      if ( current_active < 0 )
      {
        current_active = 0 ;
      }
      
      bgp -> active_time [ prefix ] = ( ( ( round - 1 ) * bgp -> active_time [ prefix ] ) + current_active ) / ( round ) ;

      bgp -> sigma [ prefix ] = sqrt ( ( ( round - 1 ) * pow ( bgp -> sigma [ prefix ], 2 ) + pow ( bgp -> active_time [ prefix ] - current_active, 2 ) ) / round ) ;

      //is this the first round?
      if ( round < 2 )
      {
        //the first round has just finished, sigma is 0, but to make some safty margin
        bgp -> sigma [ prefix ] = 1 ;
      }

      //in this implementation adaptiveMRAI is integer, but maybe I can change granularity to corespond to MRAI interval granularity
      bgp -> adaptiveMRAI [ prefix ] = bgp -> active_time [ prefix ] + 3 * bgp -> sigma [ prefix ] ;

      //adaptive must be at least 1, or next timer
      if ( bgp -> adaptiveMRAI [ prefix ] < 1 )
      {
        bgp -> adaptiveMRAI [ prefix ] = 1 ;
      }

      //in next two lines I round adapitveMRAI to the first greater multiple of granularity
      //e.g. adaptive = 1.66, granularity=0.25 => (int)(1.36/0.25)=(int)(6.64)=6 => (6+1)*0.25=1.75
      bgp -> adaptiveMRAI [ prefix ] = ( int ) ( bgp -> adaptiveMRAI [ prefix ] / bgp -> granularity ) ;
      bgp -> adaptiveMRAI [ prefix ] = ( bgp -> adaptiveMRAI [ prefix ] + 1 ) * bgp -> granularity ;
      
      next_timer = offset + ( ( int ) ( bgp -> adaptiveMRAI [ prefix ] ) ) ;
      if ( next_timer >= 30 )
      {
        next_timer -= 30 ;
      }
      
    }

    //the next_timer is not determined (there is no timer [bgp -> number_of_timers], i.e. 30), but if there are msgs to be sent,
    //the next_timer will be recalculated. If no msg is sent than next_timer will be calculated when a msg is sent from new_try_update_sent
    bgp -> prefix_timer [ prefix ] = bgp -> number_of_timers * bgp -> granularity ;

    //I want to know which round just finished for this prefix
    bgp -> round [ prefix ] ++ ;
    bgp -> idle_time [ prefix ] = 0 ;

  }

  //this round is over, so BGP has to erase all prefixes
  //Of course, if there are msgs to be sent, prefixes will be added 
  prefixes . clear ( ) ;
      
  //for each peer I first check which prefixes use this timer. These prefixes are put in the current_prefixes, and also deleted
  //from peer_timer.
  //Second, I check are there any waiting_adv for the prefixes in the current_prefixes. The wainting_adv are sent, and their prefixes
  //are placed in peer_timer (for these prefixes timer is used again)
  for (int i = 0 ; i < bgp -> nbs_ . size ( ) ; i++ )
  {
    PeerEntry * peer = bgp -> nbs_[i];

    //Maybe I don't need peer_prefix at all. Without them I will have to go through all elements of the waiting_adv
    //Who knows, maybe that is better solution, especial for small number of prefixes, for highly connected networks
    //For now I will leave the solution with peer_prefix because it seems to be more general

    //copy all prefixes for this peer in the current_prefixes
    current_prefixes = peer_prefix [ peer ] ;
    //we use current prefixes for sending msg from waiting_adv, and we will add newely send prefixes in peer_prefix
    peer_prefix [ peer ] . clear ( ) ;
                                                                                                                                                                   
    map < unsigned int, AdvPair > :: iterator i ;
    set < unsigned int > :: iterator curr_prefix ;
    for ( curr_prefix = current_prefixes . begin ( ) ; curr_prefix != current_prefixes . end ( ) ; curr_prefix ++ )
    {
    
      i = peer -> waiting_adv . find ( *curr_prefix ) ;

      prefix = *curr_prefix ;

      if ( i != peer -> waiting_adv . end ( ) )
      {

        //it's not needed anymore since the next timer saved in prefix_timer [ prefix ]
        //also, I don't need to check wheter is this msg associated with this timer
        //it MUST be associated since is in current_prefixes
        //BGP has to check whether is waiting msg associtated with this timer
          AdvPair pair =  i -> second;
          Route * waitingrte = pair.rte;
          if ( waitingrte == NULL)
          {
            peer->waiting_adv.erase( i );
            continue;
          }
        
          // advertise the waiting route
          UpdateMessage * upmsg = new UpdateMessage(bgp->nh, waitingrte);

          bgp->send(upmsg, peer);
          bgp->reset_timer(peer, BGP_Timer::KEEPALIVE);
          peer->waiting_adv.erase( i );

          //when adaptiveMRAI is used then next_timer is calculated as above
          //when is not used, next_timer is this timer, and next_timer is offset of that timer
          bgp -> reusable_timer [ next_timer ] -> peer_prefix [ peer ] . insert ( prefix ) ;
          bgp -> reusable_timer [ next_timer ] -> prefixes . insert ( prefix ) ;
          //next_timer that is going to be used for this prefix
          bgp -> prefix_timer [ prefix ] = next_timer ;

      }
    }

    //end of processing for this peer, we need to erase all prefixes from current
    current_prefixes . clear ( ) ;
  } //end of peers

  resched ( bgp -> number_of_timers * bgp -> granularity ); //default is 30 sec
}

  // ----- reusableMRAItimer.reset --------------------------------------------- //
  /**
   * Force to reset the timer,
   */
void reusableMRAItimer::reset()
{
	if (status_ == TIMER_PENDING)
  {
		cancel();
  }
	status_ = TIMER_IDLE;
}

